# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.ConnectionHandler do
  @moduledoc """
  This module is responsible for handling incoming packets and sending outgoing packets. It keeps track of what state the connection is in and which game should
  receive the packets.
  """
  alias Amethyst.Minecraft.Write
  require Logger

  @enforce_keys [:handler, :socket]
  defstruct [
    :handler,
    :socket,
    version: 0,
    connection_state: Amethyst.ConnectionState.Handshake,
    game: nil,
    encryption_state: nil,
    decryption_state: nil,
    compression_threshold: nil,
    authenticated: false,
    position: {0, 0, 0},
    player_state: %{}
  ]
  @type t :: %__MODULE__{
    handler: pid(),
    socket: :gen_tcp.socket(),
    version: 0 | 767,
    connection_state: Amethyst.ConnectionState.t(),
    game: nil | Amethyst.GameCoordinator.Game,
    encryption_state: nil | :crypto.crypto_state(),
    decryption_state: nil | :crypto.crypto_state(),
    compression_threshold: nil | non_neg_integer(),
    authenticated: boolean(),
    position: {x :: number(), y :: number(), z :: number()},
    player_state: map()
  }

  @type operation :: (__MODULE__.t() -> __MODULE__.t())
  defmodule Operations do
    @moduledoc """
    This module includes generators for common operations
    which the `ConnectionHandler` should perform.
    """
    alias Amethyst.ConnectionHandler

    @spec put(key :: atom(), value :: any()) :: ConnectionHandler.operation()
    def put(key, value) do
      fn state ->
        Map.put(state, key, value)
      end
    end

    @spec put_ps(key :: any(), value :: any()) :: ConnectionHandler.operation()
    def put_ps(key, value) do
      fn state ->
        Map.put(state, :player_state, Map.put(state.player_state, key, value))
      end
    end

    @spec send_packet(%{packet_type: atom()}) :: ConnectionHandler.operation()
    def send_packet(packet) do
      fn state ->
      try do
        data = Amethyst.ConnectionState.serialize(packet, state)

        container = if state.compression_threshold == nil do
          # Packet ID is included in data
          Write.varint(byte_size(data)) <> data
        else
          threshold = state.compression_threshold
          data_length = byte_size(data)
          if data_length >= threshold do
            compressed = Write.varint(data_length) <> :zlib.compress(data)
            Write.varint(byte_size(compressed)) <> compressed
          else
            compressed = Write.varint(0) <> data
            Write.varint(byte_size(compressed)) <> compressed
          end
        end

        encrypted = if state.encryption_state == nil do
          container
        else
          state.encryption_state |> :crypto.crypto_update(container)
        end
        :gen_tcp.send(state.socket, encrypted)
      rescue
        e ->
          Logger.error("Error sending packet #{inspect(packet)}: #{Exception.format(:error, e, __STACKTRACE__)}")
      end
      end
    end

    @spec disconnect(reason :: String.t()) :: ConnectionHandler.operation()
    def disconnect(reason) do
      fn state ->
        func = Amethyst.ConnectionState.disconnect(state, reason) |> send_packet()
        func.(state)
      end
    end
  end

  @spec child_spec(:gen_tcp.socket()) :: Supervisor.child_spec()
  def child_spec(socket) do
    %{
      id: __MODULE__,
      start: {__MODULE__, :start, [socket]}
    }
  end

  @spec start(:gen_tcp.socket()) :: no_return()
  def start(socket) do
    {:ok, spawn(fn ->
      Process.set_label("ConnectionHandler for #{inspect(socket)}")
      loop(%__MODULE__{socket: socket, handler: self()})
    end)}
  end

  @spec start_link(:gen_tcp.socket()) :: no_return()
  def start_link(socket) do
    {:ok, spawn_link(fn ->
      Process.set_label("ConnectionHandler for #{inspect(socket)}")
      loop(%__MODULE__{socket: socket, handler: self()})
    end)}
  end

  @doc """
  Informs the `handler` that a packet was received and that it should handle it.
  """
  @spec packet(handler :: pid(), id :: pos_integer(), data :: binary()) :: :ok
  def packet(handler, id, data) do
    send(handler, {:packet, id, data})
    :ok
  end

  @doc """
  Informs the `handler` that its client's socket closed the connection and that it
  should proceed with cleaning it up.
  """
  @spec close(handler :: pid()) :: :ok
  def close(handler) do
    send(handler, :closed)
    :ok
  end

  @doc """
  Runs `ops` on the handler's state. All operations in `ops` will be ran sequentially
  and will not be interrupted by another operation.
  """
  @spec run(handler :: pid(), ops :: [operation()]) :: :ok
  def run(handler, ops) do
    send(handler, {:run, ops})
  end

  @doc """
  Runs a function on the handler's state and returns the result. This cannot update the state,
  this may only be used to get data.
  """
  @spec get(handler :: pid(), f :: (__MODULE__.t() -> item)) :: item when item: any()
  def get(handler, f) do
    nonce = :rand.bytes(4)
    send(handler, {:get, self(), f, nonce})
    receive do
      {:respond, ret, ^nonce} -> ret
    end
  end

  @spec run_ops([operation()], __MODULE__.t()) :: {state :: __MODULE__.t(), successes :: [boolean()]}
  defp run_ops(ops, state) do
    Enum.reduce(Enum.with_index(ops), {state, []}, fn {op, i}, {state, sacc} ->
      try do
        {op.(state), [true | sacc]}
      rescue
        e ->
          Logger.error("Operation error in op #{i}, continuing regardless: #{Exception.format(:error, e, __STACKTRACE__)}")
          {state, [false | sacc]}
      end
    end)
  end

  @spec loop(__MODULE__.t()) :: no_return()
  defp loop(state) do
    receive do
      :closed ->
        # Socket is closed, handle our side of the disconnect
        Process.exit(self(), :normal)
      {:packet, id, data} when is_integer(id) and id >= 0 and is_binary(data) ->
        {state, successes} = run_ops(handle_packet(id, data, state), state)
        if !Enum.all?(successes) do
          Logger.error("Failed to run operations of packet #{inspect(id, base: :hex)}")
        end
        loop(state)
      {:run, ops} ->
        {state, successes} = run_ops(ops, state)
        if !Enum.all?(successes) do
          Logger.error("Failed to run requested operations.")
        end
        loop(state)
      {:get, from, op, nonce} ->
        send(from, {:respond, op.(state), nonce})
        loop(state)
    end
  end

  defp chunk_pos(x, z) do
    {floor(round(x) / 16.0), floor(round(z) / 16.0)}
  end

  defp visible_chunks_from(x, z, view_distance) do
    (x - view_distance - 3 .. x + view_distance + 3) |> Enum.flat_map(fn ix ->
      (z - view_distance - 3 .. z + view_distance + 3) |> Enum.map(fn iz ->
        {ix, iz}
      end)
    end)
  end

  defp process_chunk(to, chunk, state) do
    import Amethyst.NBT.Write
    alias Amethyst.Minecraft.Write

    chunk_array = Amethyst.Game.chunk(Map.get(state, :game), chunk)
    {cx, cz} = chunk

    # TODO: Actually do heightmaps
    # TODO: Doing all this processing could be at home somewhere else, as here it's
    # not version-agnostic here
    heightmaps = compound(%{})

    data = Enum.chunk_every(chunk_array, 16, 16, 0) # 0 -> air
      |> Enum.reduce("", fn chunk_section, acc ->
        blocks_and_lights = chunk_section |> List.flatten()
        block_count = blocks_and_lights |> Enum.filter(fn {bs, _, _} -> bs != 0 end) |> length

        # Put together the palette
        unique_blocks = MapSet.new(blocks_and_lights |> Enum.map(fn {bs, _, _} -> bs end))
        min_bpe = MapSet.size(unique_blocks) |> :math.log2() |> ceil()

        paletted_container_data = case min_bpe do
          0 ->
          # SINGLE VALUED
            Write.ubyte(0) <>
            Write.varint(MapSet.to_list(unique_blocks) |> List.first()) <>
            Write.varint(0) # No data, empty pallette

          min_bpe when min_bpe in 1..8 ->
          # INDIRECT
            # Minimum bpe accepted by minecraft is 4
            bpe = max(min_bpe, 4)
            palette = MapSet.to_list(unique_blocks) |>
              Enum.with_index() |>
              Map.new(fn {v, i} -> {v, i} end)
            paletted_blocks = blocks_and_lights |>
              Enum.map(fn {bs, _, _} -> Map.get(palette, bs) end)
            paletted_data = long_aligned_bit_string_reduce(paletted_blocks, bpe,
              fn bs -> <<bs::signed-integer-big-size(bpe)>> end)
            Write.ubyte(bpe) <>
            Write.varint(map_size(palette)) <>
            (Enum.sort(palette, fn {_k1, v1}, {_k2, v2} -> v1 < v2 end) |>
              Enum.reduce("", fn {k, _v}, acc -> acc <> Write.varint(k) end)) <>
            Write.varint(floor(bit_size(paletted_data) / 64)) <>
            paletted_data

          _ ->
          # DIRECT
            data = long_aligned_bit_string_reduce(blocks_and_lights, 15,
              fn {bs, _sl, _bl} -> <<bs::signed-integer-big-size(15)>> end)
            Write.ubyte(15) <>
            Write.varint(floor(bit_size(data) / 64)) <>
            data
        end

        acc <> Write.short(block_count) <> paletted_container_data <>
        <<0::8, 0::8, 0::8>> # TODO: This should be biome data
      end)

  sky_light_sections = Enum.chunk_every(chunk_array, 16, 16, 0)
      |> Enum.map(fn chunk_section ->
        chunk_section |> List.flatten() |> Enum.map(fn {_bs, sl, _bl} -> sl end)
      end)
  non_empty_sky_light_sections = sky_light_sections |> Enum.filter(fn section -> !Enum.all?(section, &(&1 == 0)) end)
  has_sky_light = [false] ++ (sky_light_sections |> Enum.map(fn section -> !Enum.all?(section, &(&1 == 0)) end)) ++ [false]
  empty_sky_light = has_sky_light |> Enum.map(&not/1)
  sky_light_array = non_empty_sky_light_sections |> Enum.map(&(Enum.reduce(&1, <<>>, fn light, acc ->
    <<acc::bitstring, light::size(4)>>
  end)))

  block_light_sections = Enum.chunk_every(chunk_array, 16, 16, 0)
  |> Enum.map(fn chunk_section ->
    chunk_section |> List.flatten() |> Enum.map(fn {_bs, _sl, bl} -> bl end)
  end)
  non_empty_block_light_sections = block_light_sections |> Enum.filter(fn section -> !Enum.all?(section, &(&1 == 0)) end)
  has_block_light = [false] ++ (block_light_sections |> Enum.map(fn section -> !Enum.all?(section, &(&1 == 0)) end)) ++ [false]
  empty_block_light = has_block_light |> Enum.map(&not/1)
  block_light_array = non_empty_block_light_sections |> Enum.map(&(Enum.reduce(&1, <<>>, fn light, acc ->
    <<acc::bitstring, light::size(4)>>
  end)))

  send(to, {:send_packet, %{
    packet_type: :chunk_data_and_update_light,
    chunk_x: cx, chunk_z: cz,
    heightmaps: heightmaps,
    data: data,
    block_entities: [],
    sky_light_mask: has_sky_light,
    block_light_mask: has_block_light,
    empty_sky_light_mask: empty_sky_light,
    empty_block_light_mask: empty_block_light,
    sky_light_arrays: sky_light_array |> Enum.map(&(%{sky_light_array: &1})),
    block_light_arrays: block_light_array |> Enum.map(&(%{block_light_array: &1}))
  }})
  :ok
  end

  defp long_aligned_bit_string_reduce(values, bpe, func) do
    values |> Enum.reduce(<<>>, fn value, acc ->
      ret = func.(value)
      next = <<acc::bitstring, ret::bitstring-size(bpe)>>
      # man i hope they dont suddenly change the size of a long
      if rem(bit_size(next), 64) + bpe > 64 do
        # gotta pad it
        <<next::bitstring, 0::big-size(64 - rem(bit_size(next), 64))>>
      else
        next
      end
    end)
  end

  @spec handle_packet(non_neg_integer(), binary(), __MODULE__.t()) :: [operation()]
  defp handle_packet(id, data, state) when is_struct(state, __MODULE__) do
    try do
      packet = Amethyst.ConnectionState.deserialize(state, id, data)
      Amethyst.ConnectionState.handle(state, packet)
    rescue
      e ->
        Logger.error("Error handling packet with ID #{inspect(id, base: :hex)}: #{Exception.format(:error, e, __STACKTRACE__)}")
        if Application.get_env(:amethyst, :release, false) do
          [Operations.disconnect(
            "§cError handling packet #{inspect(id, base: :hex)}:\n#{Exception.format(:error, e, __STACKTRACE__)}"
            )]
        else
          []
        end
    end
  end
end
